
/*=============================================================================
 cdf.c                              A.Lambert le 12/04/03 revu le 05/05/03

 "cumulative distribution function"

 cc -O2 -Wall -c cdf.c -I../include ;
 ar -rs ../lib/librnaIV.a cdf.o ;

=============================================================================*/

#include "rnaIV.h"

Histo  GetHistCdf(Histo hist);
Histo  GetAscHistCdf(Histo hist);
double Interpol(Histo cdf, double x);

/*=============================================================================
 GetHistCdf(): Retourne une structure 'Histo' donnant la "cdf" de l'histogramme
               'hist'.
	       Chaque element de 'cdf.vals' donne l'aire de la partie de 'hist'
	       situee a droite de l'element considere.
	       Si l'integrale de 'hist' est normalisee a 1, les valeurs du tab
	       leau decroissent de 1 a 0.
=============================================================================*/

Histo GetHistCdf(Histo hist)
{
  Map    map;
  Histo  cdf;
  int    i;
  double dh;

  if (hist.bins < 2) {
      fprintf(stderr, "GetHistCdf: too few elements in array, exit..\n");
      exit(0);
  }

  map = MapSetup(hist.hmin, hist.hmax, 0, hist.bins - 1);
  dh = 1.0 / map.ratio;          /* largeur d'un intervalle de l'histogramme */

  cdf.hmin = hist.hmin;
  cdf.hmax = hist.hmax;
  cdf.bins = hist.bins;
  cdf.samples = hist.samples;

  if ((cdf.vals = (double *) malloc(cdf.bins * sizeof(double))) == NULL) {
      fprintf(stderr, "GetHistCdf: allocation failure, exit..\n");
      exit(1);
  }

  cdf.vals[cdf.bins-1] = hist.vals[cdf.bins-1];

  for (i = cdf.bins - 2; i >= 0; i--)
      cdf.vals[i] = cdf.vals[i+1] + hist.vals[i];

  for (i = 0; i < cdf.bins; i++)  cdf.vals[i] *= dh;

  return cdf;
}
/*=============================================================================
 GetAscHistCdf(): Retourne une structure 'Histo' donnant la "cdf" ascendante de
                  l'histogramme 'hist'.
	          Chaque element de 'cdf.vals' donne l'aire de la partie de
		  'hist' situee a gauche de l'element considere.
	          Si l'integrale de 'hist' est normalisee a 1, les valeurs du
		  tableau croissent de 0 a 1.
		  variante de la fonction 'GetHistCdf'.
=============================================================================*/

Histo GetAscHistCdf(Histo hist)
{
  Map    map;
  Histo  asccdf;
  int    i;
  double dh;

  if (hist.bins < 2) {
      fprintf(stderr, "GetAscHistCdf: too few elements in array, exit..\n");
      exit(0);
  }

  map = MapSetup(hist.hmin, hist.hmax, 0, hist.bins - 1);
  dh = 1.0 / map.ratio;          /* largeur d'un intervalle de l'histogramme */

  asccdf.hmin = hist.hmin;
  asccdf.hmax = hist.hmax;
  asccdf.bins = hist.bins;
  asccdf.samples = hist.samples;

  if ((asccdf.vals = (double *) malloc(asccdf.bins * sizeof(double))) == NULL) {
      fprintf(stderr, "GetAscHistCdf: allocation failure, exit..\n");
      exit(1);
  }

  asccdf.vals[0] = hist.vals[0];

  for (i = 1; i < asccdf.bins; i++)                /* accumulation des poids */
      asccdf.vals[i] = asccdf.vals[i-1] + hist.vals[i];

  for (i = 0; i < asccdf.bins; i++)  asccdf.vals[i] *= dh;

  return asccdf;
}
/*=============================================================================
 Interpol(): Retourne la valeur obtenue par interpolation lineaire des elements
             du tableau contenu dans 'cdf' correspondant a l'abscisse 'x' entree
	     en argument.
             L'interpolation est prolongee inferieurement et superieurement par
             les valeurs initiale et finale du tableau pointe par 'cdf.vals'.
	     note (26/09/06):
	     l'argument 'x' est diminue d'une petite quantite (egale au quart
	     de la largeur d'un intervalle de l'histogramme), ceci est utile
	     lorsque l'histogramme possede peu de valeurs, et par consequent
	     des fortes discontinuites.
=============================================================================*/

double Interpol(Histo cdf, double x)
{
  Map    map;
  int    Xo;
  double dh, xo, E;

  map = MapSetup(cdf.hmin, cdf.hmax, 0, cdf.bins - 1);
  dh = 1.0 / map.ratio;          /* largeur d'un intervalle de l'histogramme */

  x -= 0.25*dh;                                  /* voir la note du 26/09/06 */

  if (x < cdf.hmin) return cdf.vals[0];                     /* prolongements */
  if (x > cdf.hmax) return cdf.vals[cdf.bins - 1];  /* si x est hors limites */

  Xo = GetXFloor(map, x);
  xo = Getx(map, Xo);

  E = cdf.vals[Xo] + (x - xo) * (cdf.vals[Xo+1] - cdf.vals[Xo]) / dh;

  return E;
}
/*===========================================================================*/
