
/*
===============================================================================
 ctrlcfgs.c                        A.Lambert le 04/12/02, revu le 30/05/05

 code dedie au calcul du nombre de configurations d'un masque sous l'option 'smp'
 ou les masques d'un tableau sont consideres independants les uns des autres).

 cc -Wall -O2 -c ctrlcfgs.c -I../include  -DDEBUG ;

 cc -Wall -O2 -c ctrlcfgs.c -I../include ;
 ar -rs ../lib/librnaIV.a ctrlcfgs.o ;

===============================================================================
*/

#include "rnaIV.h"

void  SmpCfgsNb(Mask *mask);
void  PrintMaskCfgsVol(Mask *mask, FILE *txt);
void  CfgMsg(int maskproc, FILE *txt);
int   Rnd(int min, int max);
short *DmpRndCfg(Mask *mask);
void  CtrlCfgs(Mask *mask, int nmask, int maskproc, char mode, FILE *txt);

/* fonctions de 'libsrc/mcfgs.c' */
void DmpCfgsNb(Mask *mask);
/* fonctions de 'libsrc/mgapslist.c' */
int  StressSize(Mask *mask, int stress_num);
int  StressVar(Mask *mask, int stress_num);
void ResetMaskGapsList(Mask *mask);
void ImportCfgInfos(Mask *mask, short *cfg);
void ExportMaskInfos(Mask *m1, Mask *m2);
/*
===============================================================================
 SmpCfgsNb(): Calcule le nombre de configurations du masque pointe par 'mask'
              sous l'option 'smp' (Static Mask Processing) ou les masques d'un
	      tableau sont consideres independants les uns des autres.
              Retourne ce nombre exprime en unites logarithmiques de base 2.
              Si ce nombre est inferieur a 'CFG_MAX_BITS', en retour 'ncfg'
              pointe sa valeur decimale sinon 0.
	      Le tableau 'mask->grp' est suppose correctement initialise (voir
	      'GetMaskGapsList()' dans 'libsrc/mgapslist.c').
	      Si mask->ngrps == 0 'mask->log2ncfg' et 'mask->ncfg' sont initia-
	      lises respectivement aux valeurs 0.0 et 1.
===============================================================================
*/
void SmpCfgsNb(Mask *mask)
{
  double log2ncfg;
  int    j, n;

  log2ncfg = 0.0;
  for (j = 0; j < mask->ngrps; j++) {
      log2ncfg += LOG2(mask->grp[j].max_gaps - mask->grp[j].min_gaps + 1.);
  }
  mask->log2ncfg = log2ncfg;
  mask->ncfg = 0;

  if (log2ncfg <= CFG_MAX_BITS)
  {
      n = 1;
      for (j = 0; j < mask->ngrps; j++)
          n *= (int) (mask->grp[j].max_gaps - mask->grp[j].min_gaps + 1);
      mask->ncfg = n;
  }
  return ;
}
/*
===============================================================================
 PrintMaskCfgsVol(): Affiche le nombre de configurations d'un masque et le vo-
                     lume occupe par le tableau des configurations.
                     La constuction effective du tableau n'est pas necessaire a
                     ce calcul.
                     Le volume est exprime en KB, si le nombre de configurations
                     depasse 'CFG_MAX_BITS' les unites logarithmiques de base 2
                     sont utilisees.
                     La sortie est effectuee dans le fichier pointe par 'txt'.
===============================================================================
*/
void PrintMaskCfgsVol(Mask *mask, FILE *txt)
{
  double v;                       /* v: volume en Koctet d'une configuration */

  v = 1.*(sizeof(Config) + 3 * (mask->nhx + mask->nst + 2) * sizeof(short));
  v /= 1024.;

  if (mask->log2ncfg <= CFG_MAX_BITS)
  {
      fprintf(txt, "config.number: %d\n", mask->ncfg);
      fprintf(txt, "config.table volume: %.1f KB\n", mask->ncfg * v);
  }

  else          /* affichage en 'float' pour prevenir un overflow sur 'ncfg' */
  {
      fprintf(txt, "config.number: %.1e\n", pow(2., mask->log2ncfg));
      fprintf(txt, "config.table volume: %.1e MB\n",
              pow(2., mask->log2ncfg - 10) * v);
  }
  return;
}
/*
===============================================================================
 CfgMsg(): Affiche vers 'txt' un commentaire, suivant la valeur de 'maskproc'.
===============================================================================
*/
void CfgMsg(int maskproc, FILE *txt)
{
  if (maskproc == DYNAMIC)
      fprintf(txt, "It is advisable to use extra search steps.\n");
  else
      fprintf(txt,"It is advisable to use multi-step search (masks).\n");

  fprintf(txt, "Please consult Erpin documentation about search strategies.\n");

  return;
}
/*
==============================================================================
 Rnd(): genere un nombre pseudo aleatoire dans [min, max], bornes incluses.
==============================================================================
*/
int Rnd(int min, int max)
{
 return  (min + (int) ((float) (max - min + 1) * rand() / (RAND_MAX + 1.0)));
}
/*
===============================================================================
 DmpRndCfg(): cree une configuration aleatoire du masque pointe par 'mask', qui
              simule le resultat d'une recherche compatible avec les contrain-
	      tes sur certains groupes de brins avec gaps (option 'dmp').
	      le tableau 'mask->gapslist' est suppose completement initialise.
	      Si mask->ngrps == 0 ('mask->grp' n'a pas ete cree) la fonction
	      retourne un pointeur NULL;
===============================================================================
*/
short *DmpRndCfg(Mask *mask)
{
  int   j, k, size, s, sum, stress_num;
  short *cfg;

  if (mask->ngrps == 0) return NULL;

  cfg = (short *) malloc(mask->ngrps * sizeof(short));

  for (j = 0; j < mask->ngrps; j++)
  {
      stress_num = mask->grp[j].stress_num;

      if (stress_num < 0) {       /* groupe non implique dans une contrainte */
          cfg[j] = Rnd(mask->grp[j].min_gaps, mask->grp[j].max_gaps);
      }
      else {
          size = StressSize(mask, stress_num);
	  sum = StressVar(mask, stress_num);
	  do
	  {
	      for (s = 0, k = j; k < j + size; k++) {
	          cfg[k] = Rnd(mask->grp[k].min_gaps, mask->grp[k].max_gaps);
		  s += cfg[k];
              }
	  } while (s != sum);
	  j += size - 1;                         /* indice du dernier groupe */
      }                                        /*implique dans la contrainte */
  }
  return cfg;
}
/*
===============================================================================
 CtrlCfgs(): Procede au controle du nombre de configurations de 'nmask' masques
             (le premier pointe par 'mask') dans l'ordre ou ils seront traites
             dans une operation de recherche.
             Les types de traitement consideres sont 'DYNAMIC' et 'STATIC',
             l'argument 'maskproc' prenant une de ces 2 valeurs.
             Dans le cas 'DYNAMIC' on opere une simulation de recherche qui ne
             necessite pas la creation effective des tableaux de configurations.
             Le comportement va, en fonction du nombre de configurations trouve
             et du volume de memoire correspondant requis, du calcul 'silencieux'
             a l'arret inconditionnel, en passant par l'affichage d'informations
             et/ou de 'Warning' (aide a l'utilisateur).
             Si l'argument 'mode' est 'V' l'information sera obligatoirement
             fournie.
             Les sorties sont dirigees sur le fichier pointe par 'txt'.
===============================================================================
*/
void CtrlCfgs(Mask *mask, int nmask, int maskproc, char mode, FILE *txt)
{
  short    *gapcfg = NULL;
  int      i, stops, warns, infos;
  char     *warnstr, *stopstr;

  warnstr = (char *) calloc(128, 1);
  stopstr = (char *) calloc(128, 1);

  if (maskproc != STATIC && maskproc != DYNAMIC) {
      fprintf(stderr, "CtrlCfgs: Unknown argument #3, exit..\n");
      exit(1);
  }

  stops = warns = infos = 0;

  if (maskproc == DYNAMIC && nmask > 1)        /* simulation de la reduction */
  {                                        /* progressive des configurations */
      gapcfg = DmpRndCfg(mask);
      ImportCfgInfos(mask, gapcfg);
      if (gapcfg != NULL) {
          free(gapcfg);
          gapcfg = NULL;
      }
      for (i = 1; i < nmask; i++)
      {
          ExportMaskInfos(mask + i - 1, mask + i);
          DmpCfgsNb(mask + i);

          if (i != nmask - 1)
          {
              gapcfg = DmpRndCfg(mask + i);
              ImportCfgInfos(mask + i, gapcfg);
	      if (gapcfg != NULL) {
                  free(gapcfg);
	          gapcfg = NULL;
              }
          }
      }
  }

  for (i = 0; i < nmask; i++)                          /* examen des masques */
  {
      if (mask[i].log2ncfg > CFG_BITS_INFO) {
      infos++;
      }
      if (mask[i].log2ncfg > CFG_BITS_WARN) {
          sprintf(warnstr, "%s#%d, ", warnstr, i+1);
          warns++;
      }
      if (mask[i].log2ncfg > CFG_MAX_BITS) {
          sprintf(stopstr, "%s#%d, ", stopstr, i+1);
          stops++;
      }
  }
                                               /* affichage des informations */
  if (infos > 0 || toupper(mode) == 'V' )
  {
      fprintf(txt, "\nMEMORY USAGE CONTROL: %d step%s to be processed\n",
              nmask, nmask > 1 ? "s" : "");
      for (i = 0; i < nmask; i++) {
          fprintf(txt, "step #%d:\n", i+1);
          PrintMaskCfgsVol(mask + i, txt);
      }
  }                            /* affichage precedant la sortie du programme */
  if (stops > 0) {
      fprintf(stderr, "\nWARNING: Too many configurations for step%s %s\n",
              stops > 1 ? "s" : "", stopstr);
      CfgMsg(maskproc, stderr);
      fprintf(stderr, "exit..\n");
      exit(1);
  }                /* affichage precedant la sortie optionnelle du programme */
  if (warns > 0) {
      fprintf(stderr, "\nWARNING: many configurations for step%s %s\n",
              warns > 1 ? "s" : "", warnstr);
      CfgMsg(maskproc, stderr);
      fprintf(stderr, "may be you prefer to kill the program..\n");
      sleep(5);
  }
              /*------------------------------------------------------------ */
	      /* restauration du contenu initial des masques avant la sortie */
              /*------------------------------------------------------------ */

  if (maskproc == DYNAMIC && nmask > 1)
      for (i = 0; i < nmask; i++) {
          ResetMaskGapsList(mask + i);
	  SmpCfgsNb(mask + i);
      }

  free(warnstr);
  free(stopstr);
  return;
}
/*
===============================================================================
*/
