
/*
===============================================================================
 dmp.c                             A.Lambert le 24/05/02 revu le 30/05/05

 Fonctions destinees a assurer la gestion dynamique des masques lors d'une ope-
 ration de recherche a plusieurs niveaux.

 cc -O2 -Wall -c dmp.c -I../include ;
 ar -rs ../lib/librnaIV.a dmp.o ;

===============================================================================
*/

#include "rnaIV.h"

void  ResetGapsLists(Context *ctxt, int bgn, int end);
int   RecMaskSearch(Context *ctxt, int levels);
int   Search(Context *ctxt, int nmask, int maskproc);

/* rappel: fonctions de 'libsrc/mgapslist.c' */

void  ResetMaskGapsList(Mask *mask);
void  ImportCfgInfos(Mask *mask, short *cfg);
void  ExportMaskInfos(Mask *m1, Mask *m2);

/*
===============================================================================
 ResetGapsLists(): Remet a leur valeur initiale l'ensemble des elements des
                   champs 'gapslist' et 'grp' des masques associes au tableau
		   de structures 'Context' pointe par 'ctxt'.
                   Plus precisement aux contextes dont les indices vont de 'bgn'
                   a 'end' INCLUS.
		   Interface de la fonction 'ResetMaskGapsList'de 'mgapslist.c'.
===============================================================================
*/
void ResetGapsLists(Context *ctxt, int bgn, int end)
{
  int  i;

  for (i = bgn; i <= end; i++) ResetMaskGapsList((ctxt + i)->mask);

  return;
}
/*
===============================================================================
 RecMaskSearch():
               generalites:
               Fonction de recherche de 'masques' operant en plusieurs etapes:

               - la 1ere, de preference sur un masque presentant peu de gaps,
               - la 2eme, qui demarre a chaque detection survenue dans la 1ere
	       etape, opere sur un masque plus structure et un intervalle reduit
	       au voisinage de la detection precedente et utilisant, afin de re-
	       duire la complexite, les parametres de la configuration trouvee a
	       l'etape precedente.
               - etc ..
               Cette organisation repond au besoin d'optimisation de la duree
               de l'operation de recherche.

               Les arguments: sequence, masques, parametres .. sont pointes par
               l'intermediaire de la structure Context 'ctxt'.
               L'argument 'levels' est le nombre (generalement 1, 2 ou 3 mais
	       eventuellement plus) de niveaux de la recherche, c'est egalement
	       le nombre de 'masques' recherches.

               note:
	       La variable 'detect' repere la position d'une detection depuis
               l'origine de la sequence visitee (et non pas depuis 'ctxt->bgn').
               La fonction retourne le nombre d'occurences du masque recherche
	       au niveau le plus eleve (d'indice 'levels - 1').
	       Les occurences sont comptees dans les champs 'ctxt->detects' qui
	       sont regulierement remis a 0, et cumulees dans 'ctxt->cumuls'.
===============================================================================
*/
int RecMaskSearch(Context *ctxt, int levels)
{
  int     i, j, k, l, lo, detect, bgn, index, TabLen, ScanLen;

  extern unsigned long long int TotalNucScans;
  static unsigned int jobscans = 0;

  if ( SetSeqContext(ctxt) == 0 )  return 0;       /* sequence trop courte ! */

  TabLen  = ctxt->scoretablen;
  ScanLen = ctxt->tabscanlen;

  for (i = k = 0; i <= ctxt->nscan; i++, k += ctxt->tabscanlen)
  {
      if (i == ctxt->nscan)                         /* phase finale atteinte */
      {
          TabLen  = ctxt->lastscoretabLen;
          ScanLen = ctxt->lastscanlen;

          if (TabLen == 0) break;
      }
                                                       /* tableaux de scores */
      GetMaskScoresTab(ctxt->data + k, TabLen, ctxt->mask, ctxt->level);

      if (ctxt->level == 0) {
          lo = 0;
          TotalNucScans += ScanLen;           /* decompte des bases traitees */
      }
      else lo = ctxt->mask->min_bgn;

      for (j = 0, l = lo; j < ScanLen; j++, l++)        /* --- recherche --- */
      {
          if ( GetMaskScore(ctxt->mask, l, ctxt->level) > ctxt->mask->threshold )
          {
              ctxt->detects++;
              ctxt->cumuls++;
              detect = ctxt->bgn + k + l;
              index = ctxt->mask->cfgindex;

              if (levels == 1) {
                  AddLink2(ctxt, detect, index, ctxt->mask->score);
              }
              else                   /* relance 'RecMaskSearch' au voisinage */
              {                                           /* de la detection */
                  Context *ctxt1 = ctxt + 1;

                  GetBgn(ctxt, detect, &bgn);
                  StoreBgn(ctxt1, bgn);
		  
		  SaveGapsList(ctxt->mask);       /* sauvegarde avant modif. */

		  ImportCfgInfos(ctxt->mask, ctxt->mask->gapscfg[index]);
		  ExportMaskInfos(ctxt->mask, ctxt1->mask);

                  DelMaskCfgTables(ctxt1->mask);
                  GetMaskCfgs(ctxt1->mask);

                  (int) RecMaskSearch(ctxt1, levels - 1);
		  
		  RestoreGapsList(ctxt->mask);                /* restauration */
              }
          }
      }
      PrintStatus(ctxt, &jobscans);
  }
      /* reinitialisation des tableaux de gaps avant la sortie au 1er niveau */
  if (ctxt->level == 0)  ResetGapsLists(ctxt, 0, levels - 1);

  return ctxt[levels - 1].detects;
}
/*
===============================================================================
 Search(): Interface des fonctions de recherche 'MaskSearch' et 'RecMaskSearch'
           pilote par le choix de traitement 'maskproc'(option '-dmp' ou '-smp'
           de 'erpin').
===============================================================================
*/
int Search(Context *ctxt, int nmask, int maskproc)
{
  int ndetects;

  if (maskproc == DYNAMIC)                /* recherche dynamique des masques */
  {                                              /* la plus souvent utilisee */
      (int) RecMaskSearch(ctxt, nmask);

      ndetects = fPrintfPostProcess2(ctxt + nmask - 1);
  }
  else if (maskproc == STATIC)             /* recherche statique des masques */
  {                                          /* convient a des motifs courts */
      (int) MaskSearch(ctxt, nmask);

      ndetects = fPrintfPostProcess(ctxt + nmask - 1);
  }
  else                                                             /* erreur */
  {
      fprintf(stderr, "Search: Invalid argument #3, exit..\n");
      exit(1);
  }

  return ndetects;
}
/*
===============================================================================
*/
