/*
==============================================================================
 lcalc.c                                       A.Lambert le 27/02/06

 fonctions de calcul lineaire, matrices, multiplication, Trace, inverse et
 fit polynomial.
 Ces fonctions seront utilisees pour construire une fonction polynomiale

 cc -O2 -Wall -c lcalc.c -I../include  ;

 ar -r ../lib/librnaIV.a lcalc.o ;

==============================================================================
*/

#include "rnaIV.h"

/*
rappel: macros de 'lcl.h'
#define ROWS(m)  ((int)m[-1][0])
#define COLS(m)  ((int)m[-1][1])
#define LENGTH(v)  ((int)v[-1])
*/

double **Mat(int nrow, int ncol);
void   FreeMat(double **m);
double *Vec(int len);
void   FreeVec(double *v);

void   Mult(double **M, double **N, double **MxN);
double Trace(double **M);
double Inv(double **A, double **invA);
double **Inverse(double **A);
double *MxV(double **m, double *v);
double Poly(double x, double *a);
void   CholDcmp(double **A, double *diag);
double *CholSl(double **A, double *diag, double *B);
double *Cholesky(double **A, double *B);
double *PolyFit(double *X, double *Y, int n);

void   fPrintMat(FILE *txt, char *comment, double **m);
void   fPrintVec(FILE *txt, char *comment, double *v);

/*
==============================================================================
 Mat(): Alloue la memoire et stocke le nombre de lignes et de colonnes.
        Par utilisation de calloc, le corps de la matrice est initialise a 0.
	Les dimensions du tableau sont enregistrees sur un espace memoire si-
	tuee dans les adresses negatives.
==============================================================================
*/
double **Mat(int nrow, int ncol)
{
  int    i, MATINFOS = 2;
  double **m;

  if ((m = (double **) malloc((nrow + 1) * sizeof(double *))) == NULL) {
      fprintf(stderr, "Mat: allocation failure 1, exit..\n");
      exit(1);
  }
  m += 1;

  if ((m[-1] = (double *) malloc(MATINFOS * sizeof(double))) == NULL) {
      fprintf(stderr, "Mat: allocation failure 2, exit..\n");
      exit(1);
  }
  m[-1][0] = (double) nrow;                           /* stocke la geometrie */
  m[-1][1] = (double) ncol;                                 /* de la matrice */

  if ((m[0] = (double *) calloc(nrow * ncol, sizeof(double))) == NULL) {
      fprintf(stderr, "Mat: allocation failure 3, exit..\n");
      exit(1);
  }
  for(i = 1; i < nrow; i++)  m[i] = m[i-1] + ncol;

  return m;
}
/*
==============================================================================
 FreeMat(): Libere la memoire, tableaux de pointeurs et de doubles.
==============================================================================
*/
void FreeMat(double **m)
{
  free(m[-1]);
  free(m[0]);
  free(m - 1);
}
/*
==============================================================================
 Vec(): Retourne un pointeur sur un tableau de doubles, initialise a 0, et dont
        la longueur 'len' est enregistree.
==============================================================================
*/
double *Vec(int len)
{
  double *v;

  if ((v = (double *) calloc(len+1, sizeof(double))) == NULL) {
      fprintf(stderr, "Vec: allocation failure, exit..\n");
      exit(1);
  }
  v += 1;                                         /* decallage de l'adresse */
  v[-1] = (double) len;         /* la taille est enregistree a l'adresse -1 */

  return v;
}
/*
==============================================================================
 FreeVec(): libere la memoire alouee a l'argument.
==============================================================================
*/
void FreeVec(double *v)
{
  free(v-1);
}
/*
==============================================================================
 Mult(): Calcule le produit de 2 matrices M et N dans cet ordre et charge le
         resultat dans les elements du tableau pointe par le 3eme argument.
         Un controle de la dimension des arguments est effectue des le debut.
         On notera que l'appel "Mult(M, N, M)" est legitime pour des matrices
	 carrees. En effet, un tableau temporaire est utilise pour stocker les
	 resultats au cours du calcul, les elements du 3eme argument ne sont
	 modifies qu'a la fin.
==============================================================================
*/
void Mult(double **M, double **N, double **MxN)
{
  int     i, j, k, dim,         /* dim: dimension commune de Def(M) et Val(N) */
          nrow1, nrow2, nrow3, ncol1, ncol2, ncol3;
  double  **m, sum;

  nrow1 = ROWS(M);
  ncol1 = COLS(M);

  nrow2 = ROWS(N);
  ncol2 = COLS(N);

  nrow3 = ROWS(MxN);
  ncol3 = COLS(MxN);

  if (ncol1 != nrow2) {
     fprintf(stderr, "\nMult: needs dim{Def(arg1)} = dim{Val(arg2)}, exit..\n");
     exit(1);
  }
  dim = COLS(M);

  if ((nrow3 != nrow1) || (ncol3 != ncol2)) {
     fprintf(stderr, "\nMult: invalid size of arg #3, exit..\n");
     exit(1);
  }

  m = Mat(nrow1, ncol2);

  for (i = 0; i < nrow1; i++)
     for (j = 0; j < ncol2; j++) {
        for (k = 0, sum = 0.0; k < dim; k++)   sum += M[i][k] * N[k][j];
        m[i][j] = sum;
     }
  for (i = 0; i < nrow1; i++)
     for (j = 0; j < ncol2; j++)  MxN[i][j] = m[i][j];

  FreeMat(m);
}
/*
==============================================================================
 Trace(): Retourne la trace (somme des elements diagonaux) de la matrice carree
          pointee par M.
==============================================================================
*/
double Trace(double **M)
{
  int     i, nrow, ncol;
  double  Tr;

  nrow = ROWS(M);
  ncol = COLS(M);

  if (nrow != ncol) {
     fprintf(stderr, "Trace(): arg. must be a square matrix, exit..\n");
     exit(1);
  }
  for (Tr = 0.0, i = 0; i < nrow; i++)  Tr += M[i][i];

  return Tr;
}
/*
==============================================================================
 Inv(): Inverse la matrice carree A et retourne son determinant.
        Les elements de la matrice inverse sont charges dans le 2eme argument
	qui doit pointer un tableau deja alloue. Un contole des dimensions des
        2 arguments est effectue.
        Si le determinant est trop proche de 0 le calcul est interrompu.
        On notera que l'appel "Inv(A, A)" est legitime;

	L'algorithme de Le Verrier permet, a partir des Traces des puissances
        d'une matrice et des "formules de Newton" de calculer les coefficients
        de son polynome caracteristique: P(s) = Det(s.I - A) ou I designe la
	matrice Identite.
        On deduit la matrice inverse de A de l'identite de Cayley-Hamilton:
        "Une matrice annule son polynome caracteristique".
        Ref: B.Demidovitch et I.Maron "Elements de calcul numerique" Editions
        MIR, Moscou.
==============================================================================
*/
#define DET_MIN  1.e-12

double Inv(double **A, double **invA)
{
  double **B = NULL, p;
  int    i, j, dim, nrow, ncol;

  nrow = ROWS(A);
  ncol = COLS(A);

  if (nrow != ncol) {
     fprintf(stderr, "Inv: arg. #1 must be a square matrix, exit..\n");
     exit(1);
  }
  dim = nrow;

  nrow = ROWS(invA);
  ncol = COLS(invA);

  if ((nrow != dim) || (ncol != dim)) {
     fprintf(stderr, "Inv: invalid size of arg #2, exit..\n");
     exit(1);
  }

  B = Mat(dim, dim);

  for (i = 0; i < dim; i++)
     for (j = 0; j < dim; j++) B[i][j] = A[i][j];

  p = - Trace(A);
  for (j = 0; j < dim; j++) B[j][j] += p ;       /* ajoute p a la diagonale */

  for (i = 2; i < dim ; i++) {
     Mult(A, B, B);
     p = - Trace(B) / (double)i;
     for (j = 0; j < dim; j++) B[j][j] += p ;    /* ajoute p a la diagonale */
  }
  p = 0.0;
  for (i = 0; i < dim; i++)
     p -= A[0][i] * B[i][0];

  if (fabs(p) < DET_MIN) {
     fprintf(stderr, "Inv: invalid argument, determinant too close to 0, exit..\n");
     exit(1);
  }
  else
     for (i = 0; i < dim; i++)
        for (j = 0; j < dim; j++)
           invA[i][j] = - B[i][j] / p;

  FreeMat(B);

  return ((dim%2 == 0 ? 1.0 : -1.0) * p);
}
#undef DET_MIN
/*
==============================================================================
 Inverse(): Variante de la fonction 'Inv' retournant un pointeur sur la matri-
            ce inversee (allouee par 'Mat'). Le determinant n'est pas retourne.
	    On notera que l'appel "A = Inverse(A)" est incorrect.
==============================================================================
*/
#define DET_MIN  1.e-12

double **Inverse(double **A)
{
  double **B = NULL, p;
  int    i, j, dim, nrow, ncol;

  nrow = ROWS(A);
  ncol = COLS(A);

  if (nrow != ncol) {
     fprintf(stderr, "Inverse: argument must be a square matrix, exit..\n");
     exit(1);
  }
  dim = nrow;
  B = Mat(dim, dim);

  for (i = 0; i < dim; i++)
     for (j = 0; j < dim; j++) B[i][j] = A[i][j];

  p = - Trace(A);
  for (j = 0; j < dim; j++) B[j][j] += p ;       /* ajoute p a la diagonale */

  for (i = 2; i < dim ; i++) {
     Mult(A, B, B);
     p = - Trace(B) / (double)i;
     for (j = 0; j < dim; j++) B[j][j] += p ;    /* ajoute p a la diagonale */
  }
  p = 0.0;
  for (i = 0; i < dim; i++)  p -= A[0][i] * B[i][0];
                                  /* controle du determinant avant division */
  if (fabs(p) < DET_MIN) {
     fprintf(stderr, "Inverse: invalid argument, determinant too close to 0, exit..\n");
     exit(1);
  }
  for (i = 0; i < dim; i++)
      for (j = 0; j < dim; j++)  B[i][j] /= (-p);

  return B;
}
#undef DET_MIN
/*
==============================================================================
 MxV(): Retourne un pointeur sur un tableau representant le produit de la ma-
        trice pointee par 'm' et du vecteur pointe par 'v'.
	La fonction controle la compatibilite des dimensions. La longueur du
	vecteur retourne est egale au nombre de lignes de la matrice.
==============================================================================
*/
double *MxV(double **m, double *v)
{
  int    i, j, nrow, ncol;
  double *mxv;

  nrow = ROWS(m);
  ncol = COLS(m);

  if (ncol != LENGTH(v)) {
      fprintf(stderr, "MxV: invalid length of argument #2, exit..\n");
      exit(1);
  }
  mxv = Vec(ncol);                 /* un tableau de 'ncol' elements mis a 0 */

  for (i = 0; i < nrow; i++)                     /* calcul de l'image 'mxv' */
      for (j = 0; j < ncol; j++)
           mxv[i] += m[i][j] * v[j];

  return mxv;
}
/*
===============================================================================
 Poly(): Retourne la valeur au point 'x' du polynome dont les coefficients sont
         les elements du tableau pointe par 'a'. a[i] est le coeff. du terme de
	 degre i, un tableau de 'n' elements donne un polynome de degre 'n-1'.
         Le tableau 'a' est suppose avoir ete cree par la fonction 'Vec'.
===============================================================================
*/
double Poly(double x, double *a)
{
  int    i, degre;
  double p;

  degre = LENGTH(a) - 1;

  for (p = a[degre], i = degre-1; i >= 0; i--)  p = p*x + a[i];

  return p;
}
/*
===============================================================================
 CholDcmp(): Realise la decomposition de Cholesky: L.L'= A ou A est une matrice
             carree symetrique et definie positive (U'.A.U > 0) et L et L' sont
	     2 matrices triangulaires.
	     Seule la partie triangulaire superieure de A est necessaire, elle
	     n'est pas modifiee, L occupe le triangle inferieur a l'exception de
	     la diagonale qui est enregistree dans le tableau pointe par 'diag'.
             L' est la matrice transposee de L.
	     ref: Num.Recipes in C, 2nd Ed. p 96
===============================================================================
*/
void CholDcmp(double **A, double *diag)
{
  int    dim, i, j, k;
  double sum;

  dim = ROWS(A);
  if (LENGTH(diag) != dim || COLS(A) != dim) {
      fprintf(stderr, "CholDcmp: incompatible args sizes, exit..\n");
      exit(1);
  }

  for (i = 0; i < dim; i++)
  {
      for (j = i; j < dim; j++)
      {
          for (sum = A[i][j], k = i-1; k >= 0; k--)
	      sum -= A[i][k] * A[j][k];
	  if (i == j) {
	      if (sum <= 0) {
	          fprintf(stderr,
		  "CholDcmp: arg. is not a positive definite matrix, exit..\n");
		  exit(1);
	      }
	      diag[i] = sqrt(sum);
	  }
	  else
	      A[j][i] = sum / diag[i];
      }
  }
  return;
}
/*
===============================================================================
 CholSl(): resoud, a l'aide des donnees resultant de la decomposition de
           Choleskyle systeme lineaire A.C=B et retourne un pointeur sur C.
           'A' et 'diag' doivent pointer les tableaux en sortie de 'CholDcmp()'.
===============================================================================
*/
double *CholSl(double **A, double *diag, double *B)
{
  int i, k, dim;
  double sum, *C;

  dim = ROWS(A);
  if (LENGTH(diag) != dim || LENGTH(B) != dim || COLS(A) != dim) {
      fprintf(stderr, "CholSl: incompatible args sizes, exit..\n");
      exit(1);
  }
  C = Vec(dim);

  for (i = 0; i < dim; i++)
  {
      for (sum = B[i], k = i-1; k >= 0; k--)
          sum -= A[i][k] * C[k];
      C[i] = sum / diag[i];
  }
  for (i = dim-1; i >= 0; i--)
  {
      for (sum = C[i], k = i+1; k < dim; k++)
          sum -= A[k][i] * C[k];
      C[i] = sum / diag[i];
  }
  return C;
}
/*
===============================================================================
 Cholesky(): Interface des fonctions 'CholDcmp' et 'CholSl' destinee a resoudre
             le systeme A.C = B ou A est une matrice carree symetrique definie
	     positive. Retourne un pointeur sur le tableau 'C' solution du sys-
	     teme.
===============================================================================
*/
double *Cholesky(double **A, double *B)
{
  double *diag, *C;

  diag = Vec(LENGTH(B));
  CholDcmp(A, diag);
  C = CholSl(A, diag, B);
  FreeVec(diag);

  return C;
}
/*
===============================================================================
 PolyFit(): Etant donnes 2 tableaux de meme longueur pointes par 'X' et 'Y' re-
            presentant les abscisses et ordonnees de ces points, la fonction
	    retourne un pointeur sur le tableau des coefficients du polynome
	    dont le degre est entre par 'degre' (degre+1 elements !) optimisant,
	    au sens des moindres carres, l'ajustage des points X,Y.
	    Les tableaux pointes par 'X' et 'Y' sont supposes avoir ete crees
	    par la fonction 'Vec', leur dimension est enregistree. De meme celui
	    retourne par la fonction.
===============================================================================
*/
double *PolyFit(double *X, double *Y, int degre)
{
  double s, **K, **T, *C, *B;
  int    i, j, k, npts;

  if (LENGTH(X) != LENGTH(Y)) {
      fprintf(stderr, "PolyFit: args #1 and #2 have != lenght, exit..\n");
      exit(1);
  }
  npts = LENGTH(X);
  B = Vec(degre + 1);
  T = Mat(degre + 1, npts);
  K = Mat(degre + 1, degre + 1);

                                /* matrice T (Vandermonde): (degre+1) x npts */

  for (j = 0; j < npts; j++)  T[0][j] = 1.0;              /* 1ere ligne de T */
  for (i = 1; i <= degre; i++)                           /* lignes suivantes */
      for (j = 0; j < npts; j++) T[i][j] = T[i-1][j] * X[j];

                                                      /* calcul du vecteur B */
  for (i = 0; i <= degre; i++) {
      for (s = 0.0, j = 0; j < npts; j++) s += Y[j] * T[i][j];
      B[i] = s / (double) npts;
  }
                                                   /* calcul de la matrice K */
  for (i = 0; i <= degre; i++)
      for (j = i; j <= degre; j++) {
          for (s = 0.0, k = 0; k < npts; k++) s += T[i][k] * T[j][k];
	  K[i][j] = s / (double) npts;
	  K[j][i] = K[i][j];                           /* matrice symetrique */
      }

  C = Cholesky(K, B);                      /* resolution du systeme: K.C = B */

  FreeMat(T);
  FreeMat(K);
  FreeVec(B);

  return C;
}
/*
==============================================================================
 fPrintMat(): Affiche dans le fichier pointe par 'txt' les elements d'une mat-
              rice cree par 'Mat' dans l'ordre usuel.
==============================================================================
*/
void fPrintMat(FILE *txt, char *comment, double **m)
{
  int  i, j, nrow, ncol;

  nrow  = ROWS(m);
  ncol  = COLS(m);

  fprintf(txt, "\n %s: %d x %d \n\n", comment, nrow, ncol);

  for (i = 0; i < nrow; i++) {
     fprintf(txt, "[");
     for (j = 0; j < ncol; j++)
        fprintf(txt, " %8.4f ", m[i][j]);
     fprintf(txt, "]\n");
  }
  return;
}
/*
==============================================================================
 fPrintVec(): Affiche dans le fichier pointe par 'txt' les elements d'un tab-
              leau cree par 'Vec'.
==============================================================================
*/
void fPrintVec(FILE *txt, char *comment, double *v)
{
  int  i, len;

  len = LENGTH(v);
  fprintf(txt, "\n %s: %d\n\n", comment, len);

  fprintf(txt, "[");
  for (i = 0; i < len; i++) fprintf(txt, " %8.4f ", v[i]);
  fprintf(txt, "]\n");

  return;
}
