
/*
===============================================================================
 mgapslist.c             A.Lambert le 24/05/05, revu le 20/10/05, le 17/04/06

 code dedie a la construction des tableaux 'gapslist' et 'grp' des masques, et
 a leur manipulation, notamment pour la communication entres masques.
 La construction de 'gapslist' comporte 3 etapes:
 a/ construction de 'gapslist' sans les contraintes
 b/ tableau des groupes
 c/ etablissement des contraintes
 L'etape c/ necessite de considerer l'ensemble des masques.

 Une copie de 'gapslist', 'gapslist-ori', est faite, qui servira a reinitialiser
 'gapslist'.

 cc -Wall -O2 -c mgapslist.c -I../include  -DDEBUG ;

 cc -Wall -O2 -c mgapslist.c -I../include ;
 ar -rs ../lib/librnaIV.a mgapslist.o ;

===============================================================================
*/

#include "rnaIV.h"

void GetMaskAtoms(Mask *mask);
void GetMaskAtomStr(Mask *mask);

void GetMGapsList(Mask *mask);
void GetMaskGapsGrp(Mask *mask);
void GetMaskGapsList(Mask *mask);

void UpdateMaskGapsGrp(Mask *mask);

void Set2MasksStress(Mask *m1, Mask *m2);
void SetMasksStress(Mask *mask, int nmask);
void SaveGapsList(Mask *mask);
void RestoreGapsList(Mask *mask);
void ResetMaskGapsList(Mask *mask);
int  StressSize(Mask *mask, int stress_num);
int  StressVar(Mask *mask, int stress_num);

void ImportCfgInfos(Mask *mask, short *cfg);
void ExportMaskInfos(Mask *m1, Mask *m2);

/*
===============================================================================
 GetMaskAtoms(): Determine dans le 'pattern' associe a un masque le 1er atome
                 et le nombre d'atomes correspondant a la longueur du masque.
                 Initialise les champs 'mask->atom' et 'mask->natom' de 'mask'.

                 C'est la determination des atomes de l'ENVELOPPE du masque.

                 Cette fonction permettra de calculer les configurations d'un
                 masque, en excluant le debut et la fin du 'pattern' qui ne
                 font pas partie du masque.
===============================================================================
*/
void GetMaskAtoms(Mask *mask)
{
  int  *m, i, min_pos1, max_pos2, atom1 = 0,
       id1, id2;                      /* identificateurs des atomes extremes */

  min_pos1 =  100000;  /* assure l'initialisation correcte de 'id1' et 'id2' */
  max_pos2 = -100000;
  id1 = id2 = 0;

              /* recherche des identificateurs des atomes extremes du masque */

  for (i = 0, m = mask->hxindex; i < mask->nhx; i++, m++)
  {
      if (mask->pattern->hlx[*m].db_bgn1 < min_pos1) {
          min_pos1 = mask->pattern->hlx[*m].db_bgn1;
          id1 = mask->pattern->hlx[*m].id;
      }
      if (mask->pattern->hlx[*m].db_bgn2 > max_pos2) {
          max_pos2 = mask->pattern->hlx[*m].db_bgn2;
          id2 = mask->pattern->hlx[*m].id;
      }

  }
  for (i = 0, m = mask->stindex; i < mask->nst; i++, m++)
  {
      if (mask->pattern->std[*m].db_bgn < min_pos1) {
          min_pos1 = mask->pattern->std[*m].db_bgn;
          id1 = mask->pattern->std[*m].id;
      }
      if (mask->pattern->std[*m].db_bgn > max_pos2) {
          max_pos2 = mask->pattern->std[*m].db_bgn;
          id2 = mask->pattern->std[*m].id;
      }
  }

     /* passage des identificateurs aux indices des atomes dans le 'pattern' */

  for (i = 0; i < mask->pattern->natom; i++)
  {
      if (id1 == mask->pattern->oatlist[i]) {     /* 1ere occurence de 'id1' */
          atom1 = i;
          mask->atom = mask->pattern->atom + atom1;
          break;
      }
  }
  for (i = mask->pattern->natom - 1; i >= 0; i--)
  {
      if (id2 == mask->pattern->oatlist[i]) { /* derniere occurence de 'id2' */
          mask->natom = i - atom1 + 1;
          break;
      }
  }

  return;
}
/*
===============================================================================
 GetMaskAtomStr(): Cree une chaine de longueur 'mask->natom' constituee des
                   caracteres '1' et '0' suivant que l'atome correspondant ap-
                   partient ou non au masque pointe par 'mask'.
                   Cette chaine sera utilisee pour le calcul des configurations
                   d'un masque.
===============================================================================
*/
void GetMaskAtomStr(Mask *mask)
{
  int  i, j, offset, *m;

  if (mask->atomstr != NULL) {
      fprintf(stderr, "GetMaskAtomStr: mask->atomstr already allocated, exit..\n");
      exit(1);
  }
  mask->atomstr = (char *) malloc(mask->natom + 1);

  offset = mask->atom - mask->pattern->atom;

  for (i = 0; i < mask->natom; i++)
  {
      switch (mask->atom[i].type) {
          case HLX1:
          case HLX2:
              mask->atomstr[i] = '0';
              for (j = 0, m = mask->hxindex; j < mask->nhx; j++, m++)
                  if (mask->pattern->hlx[*m].id == mask->pattern->oatlist[offset+i])
                  {
                      mask->atomstr[i] = '1';   /* l'atome appart. au masque */
                      break;
                  }
              break;
          case STD:
             mask->atomstr[i] = '0';
             for (j = 0, m = mask->stindex; j < mask->nst; j++, m++)
                  if (mask->pattern->std[*m].id == mask->pattern->oatlist[offset+i])
                  {
                      mask->atomstr[i] = '1';   /* l'atome appart. au masque */
                      break;
                  }
              break;
           default: break;
      }
  }
  mask->atomstr[mask->natom] = '\0';

  return;
}
/*
===============================================================================
 GetMGapList(): en entree un pointeur sur un masque, la fonction cree le tableau
                'mask->gapslist' destine a gerer les configurations possibles du
	        masque. Il contient les elements necessaires a la transmission
		des informations acquises d'un niveau de recherche au suivant
		(d'un masque a l'autre).

 Le tableau cree possede autant de colonnes (plus deux) que la region contient de
 brins possedant des gaps. La largeur COMMUNE aux tableaux des differents masques
 est un element central dans la transmission des informations (des tableaux qui
 manipulent le meme ensemble de brins representes par les colonnes).

 Tous les brins dont il est question ici contiennent des gaps.
 Les colonnes correspondent aux brins de la REGION consideree.
 Les "groupes de brins" sont constitues des brins mitoyens, appartenant a la
 region consideree et a l'enveloppe du masque. Ceux contenant plusieurs brins
 appartiennent a l'enveloppe du masque mais pas au masque lui-meme (trous).

 AT_INDEX    0  indices des brins contenant des gaps de la REGION consideree
 GR_INDEX    1  indices des groupes, -1 hors l'enveloppe du masque
 STATUS      2  chaque col. contient le nb. de brins du groupe, 0 hors masque
 AT_MINGAPS  3  nombre minimal de gaps de chaque brin
 AT_MAXGAPS  4  nombre maximal de gaps de chaque brin
 GR_MINGAPS  5  nombre minimal de gaps de chaque groupe de brins (cumul)
 GR_MAXGAPS  6  nombre maximal de gaps de chaque groupe de brins (cumul)
 STR_INDEX   7  indice de la contrainte associee, sinon -1
 MGAPLIST_H  8  nbre de lignes de ce tableau

 L'avant derniere colonne joue un role particulier, elle contient des informations
 globales:
 AT_INDEX    0  nombre de brins avec gaps de la region
 GR_INDEX    1  nombre de groupes des brins de l'enveloppe du masque
 STATUS      2  mis a 0, destine a contenir le nombre de relations liant les
                brins d'un meme groupe, ceci permettra en particulier d'acceder
                facilement au parametre donnant la hauteur du tableau (laquelle
		devrait varier d'un masque a l'autre. ex: pas de relations pour
		le masque de premier niveau).
 STR_INDEX   7  l'information precedente y est dupliquee.

 La derniere colonne servira a stocker les nombres de gaps associes a une contrainte
 pour les lignes appartenant a l'extention dont il est question dans ce qui suit.
 Dans une autre fonction analysant les relations entre masques le tableau se
 verra etendu par l'adjonction des lignes explicitant chacune de ces relations.
 et permettant de donner une estimation du nombre de configurations attendues
 dans le mecanisme de "dynamic mask precessing" (dmp).

 On pourra continuer de calculer a la fin de la presente fonction le nombre de
 configurations associe au "static mask processing" (GetMaskCfgsNb(mask))

 Si aucun brin de la REGION ne contient de gaps, cas dans lequel
 mask->pattern->nvarst == 0, 'mask->gapslist' n'est pas cree (mis a NULL).
 'mask->ngrps' represente le nombre de groupes de brins avec gaps du masque, il
 peut etre nul.

 Une copie 'gapslist_ori' est effectuee, qui comme 'gapslist', sera eventuelle-
 ment modifiee dans 'Set2MasksStress'.
===============================================================================
*/
void GetMGapsList(Mask *mask)
{
  int     i, j, k, g, width, nvarst, sum, indx,
          leftindex, rightindex, leftcol, rightcol;
  Pattern *region;

  region = mask->pattern;
  width = region->nvarst;

  if (mask->gapslist != NULL)   FreesMat(mask->gapslist);

  if (width == 0) {                 /* pas de brins avec gaps dans la REGION */
      mask->ngrps = 0;                                      /* a fortiori .. */
      mask->nstress = 0;
      mask->ncfg = 1;
      mask->gapslist = NULL;                    /* le tableau n'est pas cree */

      return;                                                      /* sortie */
  }
                  /*=========================================================*/
                  /*   creation des tableaux  'gapslist' et 'gapslist_ori'   */
                  /*=========================================================*/
/*
  if (mask->gapslist != NULL)                      deja fait qques lignes avant
      FreesMat(mask->gapslist);
*/
  mask->gapslist = sMat(MGAPLIST_H, width + 2);                /* allocation */
  FillsMat(mask->gapslist, MGAPLIST_H, width + 2, 0);      /* et mise a zero */

                                       /* le tableau 'gapslist_ori' est cree */
  if (mask->gapslist_ori != NULL)
      FreesMat(mask->gapslist_ori);

  mask->gapslist_ori = sMat(MGAPLIST_H, width + 2);

      /*=====================================================================*/
      /* traitement du cas ou le masque ne contient pas de brins avec gaps.  */
      /* les lignes STATUS, GR_MINGAPS et GR_MAXGAPS restent a 0.            */
      /*=====================================================================*/

  for (i = nvarst = 0; i < mask->natom; i++)  /* compte les atomes avec gaps */
      if (mask->atom[i].max_gaps > 0) nvarst++;/* dans l'enveloppe du masque */

  if (nvarst == 0)                           /* pas de brin avec gap detecte */
  {
      mask->ngrps = 0;
      mask->ncfg = 1;

      for (j = 0; j < width; j++)  mask->gapslist[GR_INDEX][j] = -1;
  }
	        /*===========================================================*/
                /* init. a -1 de la ligne STR_INDEX, indices des contraintes */
                /*===========================================================*/

  for (j = 0; j < width; j++)
      mask->gapslist[STR_INDEX][j] = -1;                 /* pour le moment.. */
  mask->nstress = 0;                                               /* idem.. */
  mask->gapslist[STR_INDEX][width] = 0;  /* recevra le nombre de contraintes */

    /*=======================================================================*/
    /* enregistrement des indices de brins avec gaps, et des nombres de gaps */
    /* dans la REGION dont depend le masque.                                 */
    /*=======================================================================*/

  for (i = j = 0; i < region->natom; i++)   /* selection des brins avec gaps */
  {
      if (region->atom[i].type == STD && region->atom[i].max_gaps > 0)
      {
          mask->gapslist[AT_INDEX][j] = (short) region->atom[i].index;
	  mask->gapslist[AT_MINGAPS][j] = 0;
          mask->gapslist[AT_MAXGAPS][j] = (short) region->atom[i].max_gaps;
	  j++;                           /* 'j' decompte les brins avec gaps */
      }
  }                                            /* en sortie on a: j == width */
  mask->gapslist[AT_INDEX][width] = width;

#ifdef DEBUG
printf("\nctrl: j=%d   width=%d\n", j, width);
#endif

  if (nvarst == 0) {
      CopysMat(mask->gapslist, MGAPLIST_H, width + 2, mask->gapslist_ori);
      return;
  }

                 /*==========================================================*/
		 /* le masque contient des brins avec gaps ...               */
                 /* initialisation des lignes STATUS, GR_MINGAPS, GR_MAXGAPS */
                 /*==========================================================*/

      /* I. determination des indices extremes des brins avec gaps du MASQUE */

  for (leftindex = i = 0; i < mask->natom; i++) {
      if (mask->atom[i].type == STD &&  mask->atom[i].max_gaps > 0) {
          leftindex = mask->atom[i].index;
          break;
      }
  }
  for (rightindex = i = mask->natom - 1; i >= 0 ; i--) {
      if (mask->atom[i].type == STD &&  mask->atom[i].max_gaps > 0) {
          rightindex = mask->atom[i].index;
          break;
      }
  }

#ifdef DEBUG
printf("ctrl: mask_left_index=%d\n", leftindex);
printf("ctrl: mask_right_index=%d\n", rightindex);
#endif

  /* II. determination des colonnes du tableau corresp. aux indices extremes */

  for (leftcol = j = 0; j < width; j++)
      if (mask->gapslist[AT_INDEX][j] == leftindex) {
          leftcol = j;
	  break;
      }
  for (rightcol = j = width - 1; j >= 0; j--)
     if (mask->gapslist[AT_INDEX][j] == rightindex) {
          rightcol = j;
	  break;
      }

  for (j = leftcol, k = 0; k < mask->natom && j <= rightindex; k++)
  {
      if (mask->atom[k].type == STD &&  mask->atom[k].max_gaps > 0)
          switch (mask->atomstr[k])   /* un brin dans l'enveloppe du masque  */
	  {
	      case '1':                          /* qui appartient au masque */
                  mask->gapslist[STATUS][j] = 1;
		  j++;
	          break;
	      case '0':                        /* n'appartient pas au masque */
	          nvarst = 0;
		  sum = 0;
                  while (mask->atomstr[k] == '0')
                  {
                      g = mask->atom[k].max_gaps;
                      if (g > 0) {
                          sum += g;                                 /* cumul */
                          nvarst++;                 /* nb de brins avec gaps */
                      }
                      k++;
                  }
		  k--;
                  if (sum > 0)   /* le nb de gaps cumules est different de 0 */
		  {
		      for (i = 0; i < nvarst; i++)  {
		          mask->gapslist[STATUS][j + i] = nvarst;
		          mask->gapslist[GR_MINGAPS][j + i] = 0;
		          mask->gapslist[GR_MAXGAPS][j + i] = (short) sum;
		      }
		      j += nvarst;
		  }
		  break;
	  }
  }
  mask->gapslist[STATUS][width] = 0;     /* recevra le nombre de contraintes */

                  /*=========================================================*/
                  /* determination de la ligne GR_INDEX, indices des groupes */
                  /*=========================================================*/

  for (j = 0, k = 0; k < width; k++)
  {
      if (mask->gapslist[STATUS][k] != 0)
      {
          indx = mask->gapslist[STATUS][k];
          for (i = 0; i < indx; i++, k++)
	      mask->gapslist[GR_INDEX][k] = j;
	  j++;
	  k--;
      }
      else mask->gapslist[GR_INDEX][k] = -1;
  }
  mask->gapslist[GR_INDEX][width] = j; /* nombre de groupes dans l'enveloppe */
                                                      /* du masque considere */
  mask->ngrps = j;

                  /*=========================================================*/
                  /* initialisation du tableau 'gapslist_ori'.               */
                  /*=========================================================*/

  CopysMat(mask->gapslist, MGAPLIST_H, width + 2, mask->gapslist_ori);

  return;
}
/*
===============================================================================
 GetMaskGapsGrp(): La fonction cree le tableau 'grp' des groupes de brins du
                   masque pointe par 'mask'.
		   Elle suppose que le tableau 'gapslist' a deja ete cree.
		   'grp' est un tableau de structures destinees a ameliorer les
		   manipulations des elements du tableau 'gapslist'.
		   Note sur la structure 'GapsGrp':
		   Les champs 'bgn', 'atom', 'size' et 'stress_num' ne seront
	           pas modifies apres leur initialisation, contrairement aux
	           champs 'min_gaps' et 'max_gaps' qui accompagnent la gestion
		   dynamique des configurations de masques. rappel:
 typedef struct {
    int bgn,      position du 1er elt du groupe parmi les col. de 'gapslist'
        atom,                      index de l'atome du 1er element du groupe
	size,                      nombre de colonnes occupees par le groupe
	min_gaps,                      nombre minimal de gaps dans le groupe
	max_gaps,                      nombre maximal de gaps dans le groupe
	stress_num;       indice 0,1,2.. d'une contrainte associee, sinon -1
    }
    GapsGrp;
 Si aucun brin du masque ne contient de gaps, auquel cas: mask->ngrps = 0, le
 tableau 'mask->grp' n'est pas cree (mis a NULL).

 rappel: 'grp' pointe un tableau relatif a l'ENVELOPPE du masque pointe par
         l'argument de la fonction.
===============================================================================
*/
void GetMaskGapsGrp(Mask *mask)
{
  int  j, width, indx, row;

  if (mask->grp != NULL) {
      free(mask->grp);
      mask->grp = NULL;
  }

  if (mask->ngrps == 0)  return;        /* le tableau n'est pas cree, sortie */

  /* sinon .. */

  width = mask->pattern->nvarst;
  mask->grp = (GapsGrp *) malloc(mask->ngrps * sizeof(GapsGrp));

  for (indx = j = 0; j < width; j++)  /* boucle sur les colonnes de gapslist */
  {
      if (mask->gapslist[GR_INDEX][j] != -1)   /* dans l'enveloppe du masque */
      {
          mask->grp[indx].bgn = j;
          mask->grp[indx].atom = mask->gapslist[AT_INDEX][j];
          mask->grp[indx].size = mask->gapslist[STATUS][j];

	  row = (mask->grp[indx].size == 1 ? AT_MINGAPS : GR_MINGAPS);
          mask->grp[indx].min_gaps = mask->gapslist[row][j];
	  mask->grp[indx].max_gaps = mask->gapslist[row + 1][j];

	  mask->grp[indx].stress_num = mask->gapslist[STR_INDEX][j];

	  j += mask->grp[indx].size;
	  j--;                    /* j passe a la derniere colonne du groupe */
	  indx++;
      }
  }
  return;
}
/*
===============================================================================
 GetMaskGapsList(): interface des 2 fonctions precedentes.
                    Cette fonction realise l'initialisation des tableaux decri-
		    vant la repartition des gaps du masque pointe par 'mask'.
		    Cette fonction sera appelee lors de la creation des masques
		    par 'GetMask' de 'libsrc/mask.c'.

		    Les relations entre groupes (contraintes) seront considerees
                    par 'Set2MasksStress' et 'SetMasksStress'.
===============================================================================
*/
void GetMaskGapsList(Mask *mask)
{
  GetMGapsList(mask);
  GetMaskGapsGrp(mask);
  SmpCfgsNb(mask);          /* init. du nb. de configurations (option 'smp') */

  return;
}
/*
===============================================================================
 UpdateMaskGapsGrp(): actualise les champs 'mingaps' et 'maxgaps' des groupes a
                      partir du contenu du tableau 'gapslist' du masque pointe
		      par 'mask'.
                      La fonction ne fait rien si le tableau 'mask->grp' n'a pas
		      pu etre cree.
                      Cette fonction sera appelee apres une modification du
 	              tableau pointe par 'gapslist' (masque pointe par 'mask').
===============================================================================
*/
void UpdateMaskGapsGrp(Mask *mask)
{
  int     j, width, indx, row;

  if (mask->ngrps == 0) return;                   /* rien a faire .., sortie */

  width = mask->pattern->nvarst;

  for (indx = j = 0; j < width; j++)  /* boucle sur les colonnes de gapslist */
  {
      if (mask->gapslist[GR_INDEX][j] != -1)   /* dans l'enveloppe du masque */
      {
	  row = (mask->grp[indx].size == 1 ? AT_MINGAPS : GR_MINGAPS);
          mask->grp[indx].min_gaps = mask->gapslist[row][j];
	  mask->grp[indx].max_gaps = mask->gapslist[row + 1][j];
	  j += mask->grp[indx].size;
	  j--;                    /* j passe a la derniere colonne du groupe */
	  indx++;
      }
  }
  return;
}
/*
===============================================================================
 Set2MasksStress(): en entree 2 masques pointes par 'm1' et 'm2' cette fonction
                    organise le passage d'informations de 'm1' a 'm2', qui se
		    traduit par l'affichage d'une ligne composee de 0 et 1:
		    un 1 par groupe de 'm2' (par GROUPE et non pas par BRIN).
		    m2->gapslist est redimmensionne par l'adjonction des lignes
		    correspondant aux contraintes relevees.
		    'm2->grp' est aussi actualise (champs 'stress_num').
		    'm1' et 'm2' sont supposes pointer 2 masques successifs
		    d'une meme region.
		    'm2->gapslist' et 'm2->gapslist_ori' sont actualises.
===============================================================================
*/
void Set2MasksStress(Mask *m1, Mask *m2)
{
  int   i, j, k, ngr1, ngr2, sz1, sz2, indx1, indx2,
        count, nstress, height, width;
  short **stresslist;

  ngr1 = m1->ngrps;
  ngr2 = m2->ngrps;

  if (ngr1 == 0 || ngr2 == 0)    return;          /* rien a faire, sortie .. */

  width = m1->pattern->nvarst;

  stresslist = sMat(ngr2, width + 2);   /* -> enregistrement des contraintes */
  FillsMat(stresslist, ngr2, width + 2, 0);

  nstress = 0;
  for (i = 0; i < ngr1; i++)        /* boucles sur les groupes du 1er masque */
  {
      sz1 = m1->grp[i].size;
      if (sz1 > 1)      /* les groupes constitues d'un seul brin sont exclus */
      {
	  indx1 = indx2 = 0;
          for (count = sz2 = j = 0; j < ngr2; j++) /* groupes du 2eme masque */
	  {
	      if (m2->grp[j].bgn == m1->grp[i].bgn)            /* meme debut */
	      {
	          indx1 = j;         /* indice du groupe dans le 2eme masque */
		  count++;
		  for (k = j; k < ngr2; k++)
		  {
		      sz2 += m2->grp[k].size;
		      if (sz2 == sz1)
		      {
		          indx2 = k;             /* indice du dernier groupe */
			  count++;
			  break;
		      }
		  }
	      }
	      if (count == 2 &&               /* les 2 extremites coincident */
	          indx2 != indx1)     /* et plusieurs groupes sont impliques */
	      {
	  	  m2->grp[indx1].stress_num = nstress;     /* enregistrement */

		  for (k = indx1; k <= indx2; k++)    /* nouvelle contrainte */
		  {                                 /* marque les groupes du */
		      stresslist[nstress][m2->grp[k].bgn] = 1;/* 2eme masque */
	          }
		                             /* stocke l'indice du groupe du */
		  stresslist[nstress][width] = i; /* 1er masque, pour y lire */
		                                  /* ulterieurement la somme */
				                      /* des gaps a utiliser */
		  nstress++;
		  break;
	      }
	  }
      }
  }
                             /* enregistre le nombre de contraintes trouvees */

  m2->nstress = m2->gapslist[STATUS][width] = nstress;

  for (j = 0; j < ngr2; j++)             /* et initialise la ligne STR_INDEX */
      m2->gapslist[STR_INDEX][m2->grp[j].bgn] = m2->grp[j].stress_num;

  m2->gapslist[STR_INDEX][width] = nstress;

  if (nstress > 0)                    /* actualise le tableau 'm2->gapslist' */
  {                                         /* ainsi que  'm2->gapslist_ori' */
      short **tmp;

      height = MGAPLIST_H + nstress;
      tmp = sMat(height, width + 2);
      CopysMat(m2->gapslist, MGAPLIST_H, width + 2, tmp);

      for (i = 0; i < nstress; i++)
          for (k = 0; k < width + 2; k++)
	      tmp[MGAPLIST_H + i][k] = stresslist[i][k];
                                           /* actualise 'm2->gapslist' apres */
      FreesMat(m2->gapslist);                           /* redimensionnement */
      m2->gapslist = sMat(height, width + 2);
      CopysMat(tmp, height, width + 2, m2->gapslist);
      FreesMat(tmp);
                                             /* actualise 'm2->gapslist_ori' */
      FreesMat(m2->gapslist_ori);                       /* redimensionnement */
      m2->gapslist_ori = sMat(height, width + 2);
      CopysMat(m2->gapslist, height, width + 2, m2->gapslist_ori);
  }
  FreesMat(stresslist);

  return;
}
/*
===============================================================================
 SetMasksStress(): en entree 1 tableau de 'nmask' masques pointe par 'mask',
                   cette fonction organise le passage d'informations entre les
		   masques successifs.
		   Elle acheve l'initialisation des parametres "permanents" des
		   masques, elle necessite et affecte l'ensemble de ceux-ci, a
		   l'exception du premier (qui est necessaire mais pas modifie).
		   Pour chaque masque un tableau 'gapslist_save' est cree, le
		   contenu reste indetermine

		   Elle sera appelee par la fonction 'GetMasks' qui, comme cette
		   fonction, travaille sur un ensemble de masques.
===============================================================================
*/
void SetMasksStress(Mask *mask, int nmask)
{
  int i, width, height;

  width = mask->pattern->nvarst;

  if (width == 0)  return;

  for (i = 1; i < nmask; i++)
      Set2MasksStress(mask + i - 1, mask + i);
                                    /* creation des tableaux 'gapslist_save' */
  for (i = 0; i < nmask; i++) {
      if (mask[i].gapslist_save != NULL) {
          fprintf(stderr,
	  "SetMasksStress: 'gapslist_save' already allocated, exit..\n");
	  exit(1);
      }
      height = MGAPLIST_H + mask[i].nstress;
      mask[i].gapslist_save = sMat(height, width + 2);
  }

  return;
}
/*
===============================================================================
 SaveGapsList(): Saisit le contenu du tableau 'gapslist' du masque pointe par
                 'mask' et le copie dans le tableau 'gapslist_save' associe au
		 meme masque.
===============================================================================
*/
void SaveGapsList(Mask *mask)
{
  int height, width;

  width = mask->pattern->nvarst;

  if (width != 0) {
      height = MGAPLIST_H + mask->nstress;
      CopysMat(mask->gapslist, height, width + 2, mask->gapslist_save);
  }
  return;
}
/*
===============================================================================
 RestoreGapsList(): Remet les elements du champ 'gapslist' du masque pointe par
                    'mask' a la valeur saisie par 'SaveGapsList'.
		    Le tableau pointe par 'mask->grp' est aussi reactualise.
===============================================================================
*/
void RestoreGapsList(Mask *mask)
{
  int height, width;

  width = mask->pattern->nvarst;

  if (width != 0) {
      height = MGAPLIST_H + mask->nstress;
      CopysMat(mask->gapslist_save, height, width + 2, mask->gapslist);
      UpdateMaskGapsGrp(mask);                 /* mise a jour de 'mask->grp' */
  }
  return;
}
/*
===============================================================================
 ResetMaskGapsList(): Remet les elements du champ 'gapslist' du masque pointe
                      par 'mask' a leur valeur originale (conservee dans le ta-
		      bleau 'gapslist_ori').
		      Le tableau pointe par 'mask->grp' est aussi reactualise.
===============================================================================
*/
void ResetMaskGapsList(Mask *mask)
{
  int height, width;

  width = mask->pattern->nvarst;

  if (width != 0) {
      height = MGAPLIST_H + mask->nstress;
      CopysMat(mask->gapslist_ori, height, width + 2, mask->gapslist);
      UpdateMaskGapsGrp(mask);                 /* mise a jour de 'mask->grp' */
  }
  return;
}
/*
===============================================================================
 StressSize(): retourne le nombre de groupes concernes par la contrainte d'in-
               dice 'stress_num' du masque pointe par 'mask'.
===============================================================================
*/
int StressSize(Mask *mask, int stress_num)
{
  int j, width, row, size = 0;

  if (stress_num < 0) return 0;      /* pour tenir compte des valeurs -1 qui */
                                    /* peuvent etre trouvees dans 'gapslist' */
  if (stress_num >= mask->nstress) {
      fprintf(stderr,"StressSize: invalid arg. #2, exit..\n");
      exit(1);
  }
  width = mask->pattern->nvarst;

  if (width != 0)
  {
      row = MGAPLIST_H + stress_num;
      size = 0;
      for (j = 0; j < width; j++)
          size += mask->gapslist[row][j];  /* accumul. de nb. egaux a 0 ou 1 */
  }
  return size;
}
/*
===============================================================================
 StressVar(): retourne le parametre donnant le nombre de gaps total de la con-
              trainte d'indice 'stress_num' du masque pointe par 'mask'.
	      Attention: 'mask->gapslist' est suppose COMPLETEMENT initialise.
===============================================================================
*/
int StressVar(Mask *mask, int stress_num)
{
  int width, row;

  if (stress_num < 0 || mask->ngrps == 0) return 0;

  if (stress_num >= mask->nstress) {
      fprintf(stderr,"StressVar: invalid arg. #2, exit..\n");
      exit(1);
  }
  width = mask->pattern->nvarst;
  row = MGAPLIST_H + stress_num;

  return (mask->gapslist[row][width + 1]);
}
/*
===============================================================================
 ImportCfgInfos(): transmet au masque pointe par 'mask' les parametres de la
                   configuration de gaps 'cfg' (les nombres de gaps trouves),
		   qui est supposee adaptee aux groupes du masque (le meme nom-
		   bre d'elements).
===============================================================================
*/
void ImportCfgInfos(Mask *mask, short *cfg)
{
  int  j, k, w, col, row;

  if (mask->ngrps == 0)  return;                   /* rien a faire, sortie.. */

  for (j = 0; j < mask->ngrps; j++)
  {
      w = mask->grp[j].size;
      row = (w == 1 ? AT_MINGAPS : GR_MINGAPS);
      col = mask->grp[j].bgn;

      for (k = 0; k < w; k++) {
          mask->gapslist[row][col + k] =
	  mask->gapslist[row + 1][col + k] = cfg[j];
      }                                                 /* actualisation des */
      mask->grp[j].min_gaps = mask->gapslist[row][col];          /* elements */
      mask->grp[j].max_gaps = mask->gapslist[row + 1][col];      /* de 'grp' */
  }
  return;
}
/*
===============================================================================
 ExportMaskInfos(): transmet au masque pointe par 'm2' les infos recueillies
                    dans celui pointe par 'm1', cela consiste en:
                    I/ copie des gaps des groupes de longueur 1 de 'm1' (la co-
		       pie peut depasser l'enveloppe du masque et etre eventu-
		       ellement utilisee par un masque suivant),
		    II/ initialisation dans 'm2' de la somme des gaps des grou-
		        pes participant aux contraintes (somme des gaps fixee).

		    'm1' et 'm2' sont supposes etre 2 masques successifs appar-
		    tenant a une meme region.
===============================================================================
*/
void ExportMaskInfos(Mask *m1, Mask *m2)
{
  int  j, k, indx, row, w, width, indx1, indx2, bgn1, bgn2;

  if (m1->ngrps == 0 || m2->ngrps == 0)  return;     /* rien a faire, sortie */

  width = m1->pattern->nvarst;

  for (j = 0; j < width; j++)                  /* copie complete de 2 lignes */
  {
      m2->gapslist[AT_MINGAPS][j] = m1->gapslist[AT_MINGAPS][j];
      m2->gapslist[AT_MAXGAPS][j] = m1->gapslist[AT_MAXGAPS][j];
  }
  for (j = 0; j < width; j++)
  {
      indx1 = m1->gapslist[GR_INDEX][j];
      indx2 = m2->gapslist[GR_INDEX][j];
      if (indx1 < 0 || indx2 < 0)
          continue;                          /* hors l'enveloppe des masques */

      bgn1 = m1->grp[indx1].bgn;
      bgn2 = m2->grp[indx2].bgn;

      if (                                          /* detection d'un groupe */
           ( w = m2->grp[indx2].size ) > 1                     /* exterieur, */
	   && m1->grp[indx1].size == w                   /* de meme longueur */
           &&  bgn2 == bgn1                            /* et de meme origine */
         )                                             /* dans les 2 masques */
      {
          for (k = j; k < w + j; k++)
	  {
              m2->gapslist[GR_MINGAPS][k] = m1->gapslist[GR_MINGAPS][k];
              m2->gapslist[GR_MAXGAPS][k] = m1->gapslist[GR_MAXGAPS][k];
          }
          j += w - 1;
      }
  }
                                       /* traitement des contraintes de 'm2' */
  for (j = 0; j < m2->nstress; j++)
  {
      row = MGAPLIST_H + j;            /* ligne allouee a la contrainte No j */
      indx = m2->gapslist[row][width];
      m2->gapslist[row][width + 1] = m1->grp[indx].max_gaps;

#ifdef DEBUG
printf("stress_param.=%d\n", m2->gapslist[row][width + 1]);
#endif
  }

  UpdateMaskGapsGrp(m2);            /* mise a jour du contenu du tableau des */
                              /* groupes 'm2->grp', la taille reste la meme  */
  return;
}
/*
===============================================================================
*/
